/*----------------------------------------------------------------------
  File    : matrix2.c
  Contents: general vector and matrix management
            functions for matrix inversion and equation system solving
  Author  : Christian Borgelt
  History : 1999.04.30 file created
            1999.05.10 functions mat_gjinv() and mat_gjdet() completed
            1999.05.11 LU decomposition functions completed
            1999.05.12 Cholesky decomposition functions completed
            1999.05.13 vector and matrix read functions completed
            1999.05.17 parameter 'det' added to function mat_gjinv()
            2001.08.17 VECTOR data type replaced by double*
            2001.10.13 covariance funcs. redesigned, mat_regress added
            2001.10.23 functions mat_addsv() and mat_var() added
            2001.10.26 equation system solution rem. from mat_regress()
            2003.11.14 bug in mat_covar() (non-negative diagonal) fixed
            2004.02.09 singular matrix checks added to mat_trinv()
            2004.02.11 function mat_tr2sym() extended (no copy option)
            2004.02.23 function mat_isovar() added (isotropic variance)
            2004.02.29 (co)variance computations improved
            2004.04.23 function mat_chlogd() added
            2004.06.30 matrix copy bug in mat_gjinv() fixed (res != mat)
            2007.11.23 functions mat_tr2sym() and mat_trtrp() improved
            2007.12.03 several functions added and others extended
            2008.03.01 reciprocals no longer stored in mat_chdecom()
            2008.03.27 semantics of associated vector simplified
            2008.04.22 weight summ. added to mat_addsv() and mat_addmp()
            2013.08.09 adapted to higher compiler warning level
            2013.08.13 adapted to preprocessor definition of DIMID
            2015.08.11 additional check added to mat_chdecom()
            2020.03.28 function mat_chdecx() added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <limits.h>
#include <float.h>
#include <assert.h>
#include "matrix.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- matrix flags --- */
#define ODDPERM    0x0002       /* odd row permutation (LU decomp.) */

/* --- functions --- */
#ifdef NDEBUG                   /* if to compile the release version */
#define DBGMSG(...)             /* suppress all debug messages */
#else                           /* if to compile the debug version */
#define DBGMSG     fprintf      /* print debug messages */
#endif

/*----------------------------------------------------------------------
  Diagonal Matrix Functions
----------------------------------------------------------------------*/

int mat_diainv (MATRIX *res, const MATRIX *mat)
{                               /* --- invert a diagonal matrix */
  DIMID  i;                     /* loop variable */
  double t;                     /* temporary buffer */

  assert(res && mat             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt));
  for (i = res->rowcnt; --i >= 0; ) {
    t = mat->els[i][i];         /* traverse the diagonal elements */
    if (!(t != 0)) {            /* if the matrix is singular, abort */
      DBGMSG(stderr, "mat_diainv: singular matrix\n"); return -1; }
    res->els[i][i] = 1/t;       /* compute reciprocal values */
  }                             /* of the diagonal elements */
  return 0;                     /* return 'ok' */
}  /* mat_diainv() */

/*----------------------------------------------------------------------
  Triangular Matrix Functions
----------------------------------------------------------------------*/

MATRIX* mat_tr2sym (MATRIX *res, const MATRIX *mat, int src)
{                               /* --- convert triang. to sym. matrix */
  DIMID  row, col;              /* loop variables */
  double *d; const double *s;   /* to traverse the matrix rows */

  assert(res && mat             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt));
  if (src == MAT_LOWER) {       /* if to use the lower triangle */
    if (res != mat) {           /* if result and source differ */
      for (row = res->rowcnt; --row >= 0; ) {
        s = mat->els[row] +(col = row +1);
        d = res->els[row] + col;
        while (--col >= 0) *--d = *--s;
      }                         /* copy the lower triangle of the */
    }                           /* source to the destination */
    for (row = res->rowcnt-1; --row >= 0; ) {
      for (d = res->els[row] +(col = res->colcnt); --col > row; )
        *--d = mat->els[col][row];
    } }                         /* copy lower to upper triangle */
  else {                        /* if to use the upper triangle */
    if (res != mat) {           /* if result and source differ */
      for (row = res->rowcnt; --row >= 0; ) {
        s = mat->els[row] +(col = res->rowcnt);
        d = res->els[row] + col;
        while (--col >= row) *--d = *--s;
      }                         /* copy the upper triangle of the */
    }                           /* source to the destination */
    for (row = res->rowcnt; --row >= 0; ) {
      for (d = res->els[row] +(col = row); --col >= 0; )
        *--d = mat->els[col][row];
    }                           /* copy upper to lower triangle */
  }
  return res;                   /* return the result matrix */
}  /* mat_tr2sym() */

/*--------------------------------------------------------------------*/

MATRIX* mat_trtrp (MATRIX *res, const MATRIX *mat, int src)
{                               /* --- transpose a triangular matrix */
  DIMID  row, col;              /* loop variables */
  double *d;                    /* to traverse the matrix rows */

  assert(res && mat             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt));
  if (src & MAT_LOWER) {        /* if to use the lower triangle */
    for (row = res->rowcnt; --row >= 0; ) {
      for (d = res->els[row] +(col = res->colcnt); --col >= row; )
        *--d = mat->els[col][row];
    } }                         /* copy lower to upper triangle */
  else {                        /* if to use the upper triangle */
    for (row = res->rowcnt; --row >= 0; ) {
      for (d = res->els[row] +(col = row +1); --col >= 0; )
        *--d = mat->els[col][row];
    }                           /* copy upper to lower triangle */
  }
  if (src & MAT_FULL)           /* if to clear opposite triangle */
    mat_crop(res, src ^ (MAT_LOWER|MAT_UPPER));
  return res;                   /* return the result matrix */
}  /* mat_trtrp() */

/*--------------------------------------------------------------------*/

double* mat_trsubst (double *res, const MATRIX *mat, int loc,
                     const double *vec)
{                               /* --- forward/backward substitution */
  DIMID        row, col;        /* loop variables */
  const double *s;              /* to traverse the matrix rows */
  double       t;               /* temporary buffer */

  assert(res && mat             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  if (loc == MAT_LOWER) {       /* if matrix is lower triangular */
    res[0] = vec[0] /mat->els[0][0];
    for (row = 0; ++row < mat->rowcnt; ) {
      s = mat->els[row]; t = vec[row];
      for (col = row; --col >= 0; )
        t -= s[col] *res[col];  /* traverse the matrix rows and */
      res[row] = t/s[row];      /* do the backward substitution */
    } }
  else {                        /* if matrix is upper triangular */
    for (row = mat->rowcnt; --row >= 0; ) {
      s = mat->els[row]; t = vec[row];
      for (col = mat->colcnt; --col > row; )
        t -= s[col] *res[col];  /* traverse the matrix rows */
      res[row] = t/s[row];      /* do the forward substitution */
    }
  }
  return res;                   /* return the result vector */
}  /* mat_trsubst() */

/*--------------------------------------------------------------------*/

int mat_trinv (MATRIX *res, const MATRIX *mat, int loc)
{                               /* --- invert a triangular matrix */
  DIMID  row, col, i;           /* loop variables */
  double *d; const double *s;   /* to traverse the matrix rows */
  double t;                     /* temporary buffer */

  assert(res && mat             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt));
  if (loc == MAT_LOWER) {       /* if matrix is lower triangular */
    for (i = res->colcnt; --i >= 0; ) {
      d    = res->els[i];       /* traverse the unit vectors */
      t    = mat->els[i][i];    /* process the diagonal element */
      if (!(t != 0)) {          /* if the matrix is singular, abort */
        DBGMSG(stderr, "mat_trinv: singular matrix\n"); return -1; }
      d[i] = 1 /t;              /* set the diagonal element */
      for (row = i; ++row < res->rowcnt; ) {
        s = mat->els[row]; t = 0;
        for (col = row; --col >= i; ) t -= d[col] *s[col];
        d[row] = t *res->els[row][row];
      }                         /* do forward substitution */
    }                           /* with lower triangular matrix */
    for (row = res->rowcnt; --row >= 0; ) {
      for (d = res->els[row] +(col = row); --col >= 0; )
        *--d = res->els[col][row];
      for (d = res->els[row] +(col = mat->colcnt); --col > row; )
        *--d = 0;               /* copy upper to lower triangle */
    } }                         /* and clear the upper triangle */
  else {                        /* if matrix is upper triangular */
    for (i = 0; i < res->colcnt; i++) {
      d    = res->els[i];       /* traverse the unit vectors */
      t    = mat->els[i][i];    /* process the diagonal element */
      if (!(t != 0)) {          /* if the matrix is singular, abort */
        DBGMSG(stderr, "mat_trinv: singular matrix\n"); return -1; }
      d[i] = 1 /t;              /* set the diagonal element */
      for (row = i; --row >= 0; ) {
        s = mat->els[row]; t = 0;
        for (col = row; ++col <= i; ) t -= d[col] *s[col];
        d[row] = t *res->els[row][row];
      }                         /* do backward substitution */
    }                           /* with upper triangular matrix */
    for (row = 0; row < res->rowcnt; row++) {
      for (d = res->els[row] +(col = res->colcnt); --col > row; )
        *--d = res->els[col][row];
      for (--d; --col >= 0; ) *--d = 0;
    }                           /* copy lower to upper triangle */
  }                             /* and clear the lower triangle */
  return 0;                     /* return 'ok' */
}  /* mat_trinv() */

/*----------------------------------------------------------------------
  Gauss-Jordan Elimination Functions

  Reference:
  W.H. Press, S.A. Teukolsky, W.T. Vetterling, and B.P. Flannery.
  Numerical Recipes in C --- The Art of Scientific Computing
  (2nd edition), pp. 36--41.
  Cambridge University Press, Cambridge, United Kingdom 1992
----------------------------------------------------------------------*/

double mat_gjdet (MATRIX *mat, int mode)
{                               /* --- Gauss-Jordan determinant */
  MATRIX *clone = NULL;         /* clone of input matrix */
  DIMID  row, col, i;           /* loop variables, index buffers */
  DIMID  pivrow, pivcol = 0;    /* row and column of the pivot elem. */
  double pivot;                 /* pivot element */
  double *p, *r;                /* pivot element row and current row */
  DIMID  *flags;                /* matrix row flags */
  double det = 1;               /* determinant of the matrix */
  double t;                     /* temporary buffer */

  assert(mat                    /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  if (mat->rowcnt <= 1)         /* if it is a 1x1 matrix, the only */
    return mat->els[0][0];      /* matrix element is the determinant */
  if (!(mode & MAT_NOCOPY)) {   /* if not to work in place */
    mat = clone = mat_clone(mat);
    if (!clone) return NAN;     /* clone the matrix */
  }                             /* and check for success */
  memset(flags = mat->map, 0, (size_t)mat->rowcnt *sizeof(DIMID));
  for (col = mat->colcnt; --col >= 0; ) {
    if (mode & MAT_PARTPIV) {   /* if to do only partial pivoting, */
      pivot = fabs(mat->els[col][col]); /* init. the pivot element */
      for (row = pivrow = pivcol = col; --row >= 0; ) {
        t = fabs(mat->els[row][col]);  /* traverse the matrix rows */
        if (t > pivot) { pivot = t; pivrow = row; }
      } }                       /* find the column pivot element */
    else {                      /* if to do full pivoting, */
      pivot = 0; pivrow = 0;    /* initialize the pivot element */
      for (row = mat->rowcnt; --row >= 0; ) {
        if (flags[row]) continue;       /* traverse the matrix rows */
        r = mat->els[row];      /* that have not been processed yet */
        for (i = mat->colcnt; --i >= 0; ) {
          if (flags[i]) continue;    /* traverse the matrix columns */
          t = fabs(r[i]);       /* that have not been processed yet */
          if (t > pivot) { pivot = t; pivrow = row; pivcol = i; }
        }                       /* find the pivot element and */
      }                         /* note its row and column */
    }
    if (!(pivot != 0))          /* if the matrix is singular, */
      return 0;                 /* its determinant is zero */
    flags[pivcol] = -1;         /* mark the pivot row/column */
    p = mat->els[pivrow];       /* and get the pivot row */
    if (pivrow != pivcol) {     /* if the pivot is not on the diag. */
      mat->els[pivrow] = mat->els[pivcol];
      mat->els[pivcol] = p;     /* swap the pivot row into place */
      det = -det;               /* and negate the determinant */
    }                           /* to take the exchange into account */
    det *= pivot = p[pivcol];   /* multiply in the pivot element */
    for (row = mat->rowcnt; --row >= 0; ) {
      if (flags[row]) continue; /* traverse the matrix rows, */
      r = mat->els[row];        /* but skip the processed ones */
      t = r[pivcol] /pivot;     /* compute the reduction factor */
      for (i = mat->colcnt; --i >= 0; )
        if (!flags[i]) r[i] -= p[i] *t;
    }                           /* traverse and reduce the matrix row */
  }                             /* by sub. a mult. of the pivot row */
  if (clone) mat_delete(clone); /* delete the created clone */
  return det;                   /* return the computed determinant */
}  /* mat_gjdet() */

/*--------------------------------------------------------------------*/

int mat_gjinv (MATRIX *res, const MATRIX *mat, int mode, double *pdet)
{                               /* --- Gauss-Jordan matrix inversion */
  DIMID  row, col, i, k;        /* loop variables, index buffers */
  DIMID  *map1, *map2;          /* exchange maps for bookkeeping */
  DIMID  pivrow, pivcol = 0;    /* row and column of the pivot elem. */
  double pivot;                 /* pivot element */
  double *p, *r;                /* pivot element row and current row */
  double det;                   /* (reciprocal of) determinant */
  double t;                     /* temporary buffer */

  assert(mat && res && res->map /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt));
  if (mat->rowcnt <= 1) {       /* if this is a 1x1 matrix */
    t = mat->els[0][0];         /* get the top left matrix element */
    if (!(t != 0)) {            /* if the matrix is singular, abort */
      DBGMSG(stderr, "mat_gjinv: singular matrix\n"); return -1; }
    res->els[0][0] = 1/t;       /* invert the only matrix element */
    return 0;                   /* and abort the function */
  }
  if (res != mat)               /* if the result differs from the */
    mat_copy(res, mat);         /* source matrix, copy the source */
  map1 = res->map;              /* get the exchange maps */
  map2 = map1 +res->rowcnt;     /* for bookkeeping */
  for (row = res->rowcnt; --row >= 0; )
    map1[row] = 0;              /* clear the row/column flags */
  det = 1;                      /* initialize the determinant */
  for (col = res->colcnt; --col >= 0; ) {
    if (mode & MAT_PARTPIV) {   /* if to do only partial pivoting, */
      pivot = fabs(res->els[col][col]); /* init. the pivot element */
      for (row = pivrow = pivcol = col; --row >= 0; ) {
        t = fabs(res->els[row][col]);  /* traverse the matrix rows */
        if (t > pivot) { pivot = t; pivrow = row; }
      } }                       /* find the column pivot element */
    else {                      /* if to do full pivoting, */
      pivot = 0; pivrow = 0;    /* initialize the pivot element */
      for (row = res->rowcnt; --row >= 0; ) {
        if (map1[row] < 0) continue;    /* traverse the matrix rows */
        r = res->els[row];      /* that have not been processed yet */
        for (i = res->colcnt; --i >= 0; ) {
          if (map1[i] < 0) continue; /* traverse the matrix columns */
          t = fabs(r[i]);       /* that have not been processed yet */
          if (t > pivot) { pivot = t; pivrow = row; pivcol = i; }
        }                       /* find the pivot element and */
      }                         /* note its row and column */
    }
    if (!(pivot != 0)) {        /* if the matrix is singular, abort */
      DBGMSG(stderr, "mat_gjinv: singular matrix\n"); return -1; }
    map1[pivcol] |= DIMID_SGN;  /* mark the pivot row/column */
    map1[col]    |= pivrow;     /* note the row and the column */
    map2[col]     = pivcol;     /* for later unscrambling */
    p = res->els[pivrow];       /* get the pivot row */
    if (pivrow != pivcol) {     /* if rows have to be exchanged */
      res->els[pivrow] = res->els[pivcol];
      res->els[pivcol] = p;     /* swap the pivot row into place and */
      det = -det;               /* update the sign of the determinant */
    }
    det *= pivot = 1/p[pivcol]; /* update the determinant */
    p[pivcol] = 1;              /* set the diagonal element */
    for (i = res->colcnt; --i >= 0; )
      p[i] *= pivot;            /* divide the pivot row by the pivot */
    for (row = res->rowcnt; --row >= 0; ) {
      if (row == pivcol)        /* traverse the matrix rows, */
        continue;               /* but skip the current (pivot) row */
      r = res->els[row];        /* get the element in the pivot */
      t = r[pivcol]; r[pivcol] = 0;      /* column and clear it */
      for (i = res->colcnt; --i >= 0; )
        r[i] -= p[i] *t;        /* reduce the matrix row by subtract. */
    }                           /* a multiple of the pivot row */
  }
  for (col = 0; col < res->colcnt; col++) {
    i = map1[col] & ~DIMID_SGN; /* traverse the exchange maps */
    k = map2[col];              /* if no exchange is necessary, */
    if (i == k) continue;       /* skip the map entries */
    for (row = res->rowcnt; --row >= 0; ) {
      r = res->els[row];        /* traverse the matrix rows */
      t = r[i]; r[i] = r[k]; r[k] = t;
    }                           /* swap the columns into place */
  }                             /* reversing the exchange order */
  if (pdet) *pdet = det;        /* set the determinant of the inverse */
  return 0;                     /* return 'ok' */
}  /* mat_gjinv() */

/*--------------------------------------------------------------------*/

int mat_gjsol (double **res, MATRIX *mat,
               double *const*vec, DIMID cnt, int mode, double *pdet)
{                               /* --- solve linear equation systems */
  MATRIX *clone = NULL;         /* clone of input matrix */
  DIMID  row, col, i, k;        /* loop variables, index buffers */
  DIMID  *map1, *map2;          /* exchange maps for bookkeeping */
  DIMID  pivrow, pivcol = 0;    /* row and column of the pivot elem. */
  double pivot;                 /* pivot element */
  double *p, *r;                /* pivot element row and current row */
  double *v;                    /* to traverse the rhs vectors */
  double det;                   /* (reciprocal of) determinant */
  double t;                     /* temporary buffer */

  assert(mat && vec && res      /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt) && (cnt >= 0));
  if (mat->rowcnt <= 1) {       /* if the matrix has size 1x1 */
    t = mat->els[0][0];         /* get the top left matrix element */
    if (!(t != 0)) {            /* if the matrix is singular, abort */
      DBGMSG(stderr, "mat_gjsol: singular matrix\n"); return -1; }
    for (i = cnt; --i >= 0;)    /* otherwise divide all numbers */
      res[i][0] /= t;           /* by the only matrix element */
    if (mode & MAT_INVERSE)     /* if to compute the inverse matrix, */
      mat->els[0][0] = 1/t;     /* invert the only matrix element */
    return 0;                   /* and abort the function */
  }                             /* if this is a 2x2 matrix */
  if (!(mode & MAT_NOCOPY)) {   /* if not to work in place */
    mat = clone = mat_clone(mat);
    if (!clone) return -1;      /* clone the matrix */
  }                             /* and check for success */
  if ((double *const*)res != vec) {  /* if source and dest. differ, */
    for (i = cnt; --i >= 0; )   /* copy the right hand side vectors */
      vec_copy(res[i], vec[i], mat->rowcnt);
  }
  map1 = mat->map;              /* get the exchange maps */
  map2 = map1 +mat->rowcnt;     /* for bookkeeping */
  for (row = mat->rowcnt; --row >= 0; )
    map1[row] = 0;              /* clear the row/column flags */
  det = 1;                      /* initialize the determinant */
  for (col = mat->colcnt; --col >= 0; ) {
    if (mode & MAT_PARTPIV) {   /* if to do only partial pivoting, */
      pivot = fabs(mat->els[col][col]); /* init. the pivot element */
      for (row = pivrow = pivcol = col; --row >= 0; ) {
        t = fabs(mat->els[row][col]);  /* traverse the matrix rows */
        if (t > pivot) { pivot = t; pivrow = row; }
      } }                       /* find the column pivot element */
    else {                      /* if to do full pivoting, */
      pivot = 0; pivrow = 0;    /* initialize the pivot element */
      for (row = mat->rowcnt; --row >= 0; ) {
        if (map1[row] & DIMID_SGN)
          continue;             /* traverse the matrix rows */
        r = mat->els[row];      /* that have not been processed yet */
        for (i = mat->colcnt; --i >= 0; ) {
          if (map1[i] & DIMID_SGN)
            continue;           /* traverse the matrix columns */
          t = fabs(r[i]);       /* that have not been processed yet */
          if (t > pivot) { pivot = t; pivrow = row; pivcol = i; }
        }                       /* find the pivot element and */
      }                         /* note its row and column */
    }
    if (!(pivot != 0)) {        /* if the matrix is singular, abort */
      DBGMSG(stderr, "mat_gjsol: singular matrix\n"); return -1; }
    map1[pivcol] |= DIMID_SGN;  /* mark the pivot row/column */
    map1[col]    |= pivrow;     /* note the row and the column */
    map2[col]     = pivcol;     /* for later unscrambling */
    p = mat->els[pivrow];       /* get the pivot row */
    if (pivrow != pivcol) {     /* if rows have to be exchanged */
      mat->els[pivrow] = mat->els[pivcol];
      mat->els[pivcol] = p;     /* swap the pivot row into place and */
      det = -det;               /* update the sign of the determinant */
      for (i = cnt; --i >= 0; ) {
        v = res[i]; t = v[pivrow];
        v[pivrow] = v[pivcol]; v[pivcol] = t;
      }                         /* swap vector elements in the */
    }                           /* same way as the matrix rows */
    det *= pivot = 1/p[pivcol]; /* update the determinant */
    p[pivcol] = 1;              /* set the diagonal element */
    for (i = mat->colcnt; --i >= 0; )
      p[i] *= pivot;            /* divide the pivot row by the pivot */
    for (i = cnt; --i >= 0; )   /* do the same with the rhs vectors */
      res[i][pivcol] *= pivot;
    for (row = mat->rowcnt; --row >= 0; ) {
      if (row == pivcol)        /* traverse the matrix rows, */
        continue;               /* but skip the current (pivot) row */
      r = mat->els[row];        /* get the element in the pivot */
      t = r[pivcol]; r[pivcol] = 0;      /* column and clear it */
      for (i = mat->colcnt; --i >= 0; )
        r[i] -= p[i] *t;        /* reduce the matrix row */
      for (i = cnt; --i >= 0;){ /* and the rhs vectors */
        v = res[i]; v[row] -= v[pivcol] *t; }
    }                           /* i.e., subtract from the row */
  }                             /* a multiple of the pivot row */
  if (mode & MAT_INVERSE) {     /* if to compute the inverse matrix */
    for (col = 0; col < mat->colcnt; col++) {
      i = map1[col] & ~DIMID_SGN; /* traverse the exchange maps */
      k = map2[col];            /* if no exchange is necessary, */
      if (i == k) continue;     /* skip the map entries */
      for (row = mat->rowcnt; --row >= 0; ) {
        r  = mat->els[row];     /* traverse the matrix rows */
        t = r[i]; r[i] = r[k]; r[k] = t;
      }                         /* swap the columns into place */
    }                           /* reversing the exchange order */
  }                             /* (unscramble the matrix columns) */
  if (pdet) *pdet = det;        /* set the determinant of the inverse */
  if (clone) mat_delete(clone); /* delete the created clone */
  return 0;                     /* return 'ok' */
}  /* mat_gjsol() */

/*----------------------------------------------------------------------
  LU Decomposition Functions

  Reference:
  W.H. Press, S.A. Teukolsky, W.T. Vetterling, and B.P. Flannery.
  Numerical Recipes in C --- The Art of Scientific Computing
  (2nd edition), pp. 43--49.
  Cambridge University Press, Cambridge, United Kingdom 1992
----------------------------------------------------------------------*/

int mat_ludecom (MATRIX *res, const MATRIX *mat)
{                               /* --- LU decompose a matrix */
  DIMID  row, col, i;           /* loop variables, index buffers */
  DIMID  pivrow;                /* row of the pivot element */
  double pivot;                 /* pivot element/row maximum */
  double *r, *c;                /* matrix row/column buffer */
  double t;                     /* temporary buffer */

  assert(mat && res && res->map /* check the function arguments */
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt)
  &&    (mat->rowcnt == mat->colcnt));
  if (res != mat)               /* if the result differs from the */
    mat_copy(res, mat);         /* source matrix, copy the source */
  for (row = res->rowcnt; --row >= 0; )
    res->map[row] = row;        /* initialize the matrix row map and */
  res->flags &= ~ODDPERM;       /* clear the permutation parity flag */
  c = res->buf;                 /* get the internal column buffer */
  for (col = 0; col < res->colcnt; col++) {
    for (row = 0; row <= col; row++) {
      r = res->els[row]; t = r[col];
      for (i = row; --i >= 0; ) t -= c[i] *r[i];
      c[row] = r[col] = t;      /* compute a column of the */
    }                           /* upper triangular matrix U */
    pivot  = fabs(c[col]);      /* initialize the pivot element */
    pivrow = col;               /* and its row index */
    for (row = res->rowcnt; --row > col; ) {
      r = res->els[row]; t = r[col];
      for (i = col; --i >= 0; ) t -= c[i] *r[i];
      r[col] = t;               /* compute a column of the */
      t = fabs(t);              /* lower triangular matrix L */
      if (t > pivot) { pivot = t; pivrow = row; }
    }                           /* update the pivot element */
    if (!(pivot != 0)) {        /* if the matrix is singular, abort */
      DBGMSG(stderr, "mat_ludecom: singular matrix\n"); return -1; }
    r = res->els[pivrow];       /* get the row of the pivot element */
    if (pivrow != col) {        /* if rows need to be exchanged */
      res->els[pivrow] = res->els[col];
      res->els[col]    = r;     /* exchange current and pivot row */
      i                = res->map[pivrow];
      res->map[pivrow] = res->map[col];
      res->map[col]    = i;     /* exchange the row map entries */
      res->flags ^= ODDPERM;    /* and toggle the parity flag */
    }                           /* of the row permutation */
    t = 1/r[col];               /* compute the pivot factor */
    for (row = res->rowcnt; --row > col; )
      res->els[row][col] *= t;  /* divide the column of the lower */
  }                             /* triangular matrix L by the pivot */
  return 0;                     /* return 'ok' */
}  /* mat_ludecom() */

/*--------------------------------------------------------------------*/

double* mat_lusubst (double *res, const MATRIX *mat, const double *vec)
{                               /* --- fw/bw subst. for LU decomp. */
  DIMID        row, col, i;     /* loop variables, first nonzero el. */
  const double *r;              /* current matrix row */
  double       *v;              /* copied and mapped vector */
  double       t;               /* temporary buffer */

  assert(res && vec && mat      /* check the function arguments */
  &&     mat->map && (mat->rowcnt == mat->colcnt));
  for (v = mat->buf +(row = i = mat->rowcnt); --row >= 0; ) {
    *--v = vec[mat->map[row]];  /* copy and map the input vector and */
    if (*v != 0) i = row;       /* find the first nonzero element */
  }
  for (row = i; ++row < mat->rowcnt; ) {
    r = mat->els[row]; t = v[row];
    for (col = row; --col >= i; ) t -= v[col] *r[col];
    v[row] = t;                 /* do forward substitution with */
  }                             /* the lower triangular matrix L */
  for (row = mat->rowcnt; --row >= 0; ) {
    r = mat->els[row]; t = v[row];
    for (col = mat->colcnt; --col > row; ) t -= res[col] *r[col];
    res[row] = t /r[row];       /* do backward substitution with */
  }                             /* the upper triangular matrix U */
  return res;                   /* return the result vector */
}  /* mat_lusubst() */

/*--------------------------------------------------------------------*/

double mat_ludet (const MATRIX *mat)
{                               /* --- compute det. from LU decomp. */
  DIMID  i;                     /* loop variable */
  double det;                   /* determinant of the matrix */

  assert(mat && mat->map        /* check the functions argument */
  &&    (mat->rowcnt == mat->colcnt));
  det = (mat->flags & ODDPERM)  /* get the initial value depending on */
      ? -1 : +1;                /* the parity of the row permutation */
  for (i = mat->rowcnt; --i >= 0; )
    det *= mat->els[i][i];      /* multiply the diagonal elements */
  return det;                   /* and return the result */
}  /* mat_ludet() */

/*--------------------------------------------------------------------*/

int mat_luinv (MATRIX *res, const MATRIX *mat)
{                               /* --- invert matrix from LU decomp. */
  MATRIX       *clone = NULL;   /* clone of LU decomposition */
  DIMID        row, col, i, k;  /* loop variables, index buffer */
  const double *p;              /* current matrix row */
  double       *v;              /* copied and mapped vector */
  double       t;               /* temporary buffer */

  assert(mat && res && mat->map /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt));
  if (res == mat) {             /* if the result matrix is identical */
    mat = clone = mat_clone(mat);  /* to the LU decomposition matrix, */
    if (!clone) return -1;      /* clone the LU decomposition matrix */
  }                             /* to make space for the inverse */
  v = res->buf;                 /* traverse the matrix columns */
  for (i = mat->colcnt; --i >= 0; ) {
    for (v += (row = i), *v = 1; --row >= 0; )
      *--v = 0;                 /* build the next unit vector */
    for (row = i; ++row < mat->rowcnt; ) {
      p = mat->els[row]; t = 0;
      for (col = row; --col >= i; ) t -= v[col] *p[col];
      v[row] = t;               /* traverse the rows of the matrix L */
    }                           /* and do the forward substitution */
    k = mat->map[i];            /* get the destination column */
    for (row = mat->rowcnt; --row >= 0; ) {
      p = mat->els[row]; t = v[row];
      for (col = mat->colcnt; --col > row; ) t -= v[col] *p[col];
      res->els[row][k] = v[row] = t /p[row];
    }                           /* traverse the rows of the matrix U */
  }                             /* and do the backward substitution */
  if (clone) mat_delete(clone); /* delete the created clone */
  return 0;                     /* return 'ok' */
}  /* mat_luinv() */

/*----------------------------------------------------------------------
  Cholesky Decomposition Functions
  (for symmetric positive definite matrices only)

  Reference:
  W.H. Press, S.A. Teukolsky, W.T. Vetterling, and B.P. Flannery.
  Numerical Recipes in C --- The Art of Scientific Computing
  (2nd edition), pp. 96--98.
  Cambridge University Press, Cambridge, United Kingdom 1992
----------------------------------------------------------------------*/

int mat_chdecom (MATRIX *res, const MATRIX *mat)
{                               /* --- Cholesky decompose a matrix */
  DIMID        row, col, i;     /* loop variables */
  const double *s;              /* to traverse the source matrix rows */
  double       *r1, *r2;        /* to traverse the result matrix rows */
  double       t, q;            /* temporary buffers */

  assert(mat && res             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt));
  for (row = 0; row < mat->rowcnt; row++) {
    s  = mat->els[row]; t = s[row];    /* traverse the matrix rows */
    r1 = res->els[row];         /* compute square of diag. element */
    for (i = row; --i >= 0; ) t -= r1[i] *r1[i];
    if (!(t > 0)) {             /* if not positive definite, abort */
      DBGMSG(stderr, "mat_chdecom: matrix not positive definite "
                     "(%d:%.16g)\n", row, t); return -1; }
    r1[row] = t = sqrt(t);      /* set the diagonal element and */
    if (!(t > 0)) {             /* if not positive definite, abort */
      DBGMSG(stderr, "mat_chdecom: matrix not positive definite "
                     "(%d:%.16g)\n", row, t); return -1; }
    q = (t > 0) ? 1/t : INFINITY;   /* compute reciprocal value */
    for (col = row; ++col < mat->colcnt; ) {
      r2 = res->els[col]; t = s[col];
      for (i = row; --i >= 0; ) t -= r1[i] *r2[i];
      r2[row] = t*q;            /* compute the off-diagonal elements */
    }                           /* (the lower triangular matrix) and */
  }                             /* store them in the result matrix */
  return 0;                     /* return 'ok' */
}  /* mat_chdecom() */          /* (only lower triangle is computed) */

/*--------------------------------------------------------------------*/

int mat_chdecx (MATRIX *res, const MATRIX *mat, double reg,
                double rmin, double rmax)
{                               /* --- extended Cholesky decomp. */
  DIMID i;                      /* loop variable */

  assert(mat && res             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt)
  &&    (res->colcnt == mat->colcnt));
  if (reg > 0.0)                /* if regularization parameter given */
    for (i = mat->rowcnt; --i >= 0; )
      mat->els[i][i] += reg;    /* add parameter to matrix diagonal */
  for (i = mat->rowcnt; --i >= 0; )
    mat->buf[i] = mat->els[i][i];
  reg = rmin;                   /* save diagonal, get reg. parameter */
  while (1) {                   /* Cholesky decomposition trial loop */
    if (mat_chdecom(res, mat) == 0)
      break;                    /* try Cholesky decomposition */
    if (reg > rmax) return -1;  /* limit regularization steps */
    for (i = mat->rowcnt; --i >= 0; )
      mat->els[i][i] = mat->buf[i] +reg;
    reg *= 2;                   /* add regularization and try again */
  }                             /* (also increase reg. parameter) */
  return 0;                     /* return 'ok' */
}  /* mat_chdecx() */

/*--------------------------------------------------------------------*/

double* mat_chsubst (double *res, const MATRIX *mat, const double *vec)
{                               /* --- fw/bw subst. for Cholesky dec. */
  DIMID  row, col;              /* loop variables */
  double *b;                    /* reciprocals of diagonal elements */
  double *s;                    /* to traverse the matrix rows */
  double t;                     /* temporary buffer */

  assert(res && vec && mat      /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  b = mat->buf;                 /* get reciprocals of diag. elements */
  for (row = mat->rowcnt; --row >= 0; ) {
    t = mat->els[row][row]; b[row] = (t > 0) ? 1/t : INFINITY; }
  for (row = 0; row < mat->rowcnt; row++) {
    s = mat->els[row]; t = vec[row];
    b[row] = (s[row] > 0) ? 1/s[row] : INFINITY;
    for (col = row; --col >= 0; )
      t -= res[col] *s[col];    /* do forward substitution with */
    res[row] = t *b[row];       /* the lower triangular matrix L */
  }
  for (row = mat->rowcnt; --row >= 0; ) {
    s = mat->els[row]; t = res[row];
    for (col = row; ++col < mat->colcnt; )
      t -= res[col] *mat->els[col][row];
    res[row] = t *b[row];       /* do backward substitution with */
  }                             /* the upper triangular matrix L^T */
  return res;                   /* (but access only lower triangle) */
}  /* mat_chsubst() */          /* and return the result vector */

/*--------------------------------------------------------------------*/

double mat_chdet (const MATRIX *mat)
{                               /* --- compute det. from Cholesky d. */
  DIMID  i;                     /* loop variable */
  double det;                   /* determinant of the matrix */

  assert(mat                    /* check the function argument */
  &&    (mat->rowcnt == mat->colcnt));
  det = mat->els[0][0];         /* multiply diagonal elements */
  for (i = mat->rowcnt; --i > 0; ) det *= mat->els[i][i];
  return det *det;              /* return the square of the product, */
}  /* mat_chdet() */            /* which is the determinant */

/*--------------------------------------------------------------------*/

double mat_chlogd (const MATRIX *mat)
{                               /* --- compute logarithm of determ. */
  DIMID  i;                     /* loop variable */
  double det = 0, t;            /* determinant of the matrix, buffer */

  assert(mat                    /* check the function argument */
  &&    (mat->rowcnt == mat->colcnt));
  for (i = mat->rowcnt; --i >= 0; ) {
    t = mat->els[i][i];         /* get the diagonal element and */
    if (!(t > 0)) return NAN;   /* check for a positive value */
    det += log(mat->els[i][i]); /* sum the logarithms of */
  }                             /* the diagonal elements */
  return det +det;              /* return twice the sum, which is the */
}  /* mat_chlogd() */           /* logarithm of the determinant */

/*--------------------------------------------------------------------*/

int mat_chinv (MATRIX *res, const MATRIX *mat)
{                               /* --- invert matrix from Cholesky d. */
  DIMID        row, col, i;     /* loop variables */
  const double *s;              /* to traverse the source matrix rows */
  double       *d, *b;          /* to traverse the dest.  matrix rows */
  double       t;               /* temporary buffer */

  assert(res && mat             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (mat->rowcnt == res->rowcnt)
  &&    (mat->colcnt == res->colcnt));
  b = mat->buf;                 /* get reciprocals of diag. elements */
  for (row = mat->rowcnt; --row >= 0; ) {
    t = mat->els[row][row]; b[row] = (t > 0) ? 1/t : INFINITY; }
  for (i = mat->colcnt; --i >= 0; ) {
    d = res->els[i];            /* traverse the unit vectors */
    d[i] = b[i];                /* and compute their originals */
    for (row = i; ++row < mat->rowcnt; ) {
      s = mat->els[row];        /* traverse the matrix rows */
      for (t = 0, col = row; --col >= i; )
        t -= d[col] *s[col];    /* do forward substitution with */
      d[row] = t *b[row];       /* the lower triangular matrix L */
    }
    for (row = mat->rowcnt; --row >= i; ) {
      t = d[row];               /* traverse the matrix columns */
      for (col = mat->colcnt; --col > row; )
        t -= d[col] *mat->els[col][row];
      d[row] = t *b[row];       /* do backward substitution with */
    }                           /* the upper triangular matrix L^T */
  }                             /* (but access only lower triangle) */
  return 0;                     /* return 'ok' */
}  /* mat_chinv() */            /* (only upper triangle is computed) */

/*--------------------------------------------------------------------*/

int mat_chinvx (MATRIX *res, const MATRIX *mat)
{                               /* --- invert matrix with Cholesky d. */
  int r = mat_chdecom(res,mat); /* Cholesky decompose the matrix */
  if (r >= 0) mat_chinv(res, res);
  return r;                     /* compute inverse from decomposition */
}  /* mat_chinvx() */

/*----------------------------------------------------------------------
  (Co)Variance/Correlation Matrix Functions
----------------------------------------------------------------------*/

MATRIX* mat_addvec (MATRIX *mat, const double *vec, double wgt)
{                               /* --- add a vector to a matrix */
  DIMID  row;                   /* loop variable */
  double *b;                    /* to traverse the buffer */

  assert(mat && vec             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  mat->weight += wgt;           /* sum the vector weight */
  for (b = mat->vec +(row = mat->rowcnt); --row >= 0; )
    *--b += vec[row] *wgt;      /* sum the weighted vector */
  return mat;                   /* return the matrix worked on */
}  /* mat_addvec() */

/*--------------------------------------------------------------------*/

MATRIX* mat_addsv (MATRIX *mat, const double *vec, double wgt)
{                               /* --- add squared vector to a matrix */
  DIMID row;                    /* loop variable */

  assert(mat && vec             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  mat->weight += wgt;           /* sum the vector weight */
  for (row = mat->rowcnt; --row >= 0; )
    mat->els[row][row] += vec[row] *vec[row] *wgt;
  return mat;                   /* sum product of vector with itself */
}  /* mat_addsv() */            /* return the matrix worked on */

/*--------------------------------------------------------------------*/

MATRIX* mat_addsvx (MATRIX *mat, const double *vec, double wgt)
{                               /* --- add squared vector to a matrix */
  DIMID  row;                   /* loop variable */
  double *b;                    /* to traverse the buffer */
  double t;                     /* temporary buffer */

  assert(mat && vec             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  mat->weight += wgt;           /* sum the vector weight */
  for (b = mat->vec +(row = mat->rowcnt); --row >= 0; ) {
    *--b += t = vec[row] *wgt;  /* sum the weighted vector */
    mat->els[row][row] += t *vec[row];
  }                             /* sum product of vector with itself */
  return mat;                   /* return the matrix worked on */
}  /* mat_addsvx() */

/*--------------------------------------------------------------------*/

MATRIX* mat_addmp (MATRIX *mat, const double *vec, double wgt)
{                               /* --- add matrix product to a matrix */
  DIMID  row, col;              /* loop variables */
  double *d;                    /* to traverse the matrix rows */
  double t;                     /* temporary buffer */

  assert(mat && vec             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  mat->weight += wgt;           /* sum the vector weight */
  for (row = mat->rowcnt; --row >= 0; ) {
    t = vec[row] *wgt;          /* get the weighted vector element */
    for (d = mat->els[row] +(col = mat->colcnt); --col >= row; )
      *--d += vec[col] *t;      /* sum the weighted matrix product */
  }                             /* of the vector with itself */
  return mat;                   /* return the matrix worked on */
}  /* mat_addmp() */

/*--------------------------------------------------------------------*/

MATRIX* mat_addmpx (MATRIX *mat, const double *vec, double wgt)
{                               /* --- add matrix product to a matrix */
  DIMID  row, col;              /* loop variables */
  double *b, *d;                /* to traverse the buffer/matrix rows */
  double t;                     /* temporary buffer */

  assert(mat && vec             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  mat->weight += wgt;           /* sum the vector weight */
  for (b = mat->vec +(row = mat->rowcnt); --row >= 0; ) {
    *--b += t = vec[row] *wgt;  /* sum the weighted vector */
    for (d = mat->els[row] +(col = mat->colcnt); --col >= row; )
      *--d += vec[col] *t;      /* sum the weighted matrix product */
  }                             /* of the vector with itself */
  return mat;                   /* return the matrix worked on */
}  /* mat_addmpx() */

/*--------------------------------------------------------------------*/

double* mat_mean (double *res, const MATRIX *mat)
{                               /* --- mean value computation */
  DIMID  i;                     /* loop variable */
  double w;                     /* temporary buffer */
  const double *s;              /* to traverse the source vector */

  assert(mat && res             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  w = mat->weight;              /* get the weight sum and compute */
  w = (w > 0) ? 1/w : 0;        /* the factor for the mean estimation */
  for (s = mat->vec +(i = mat->rowcnt); --i >= 0; )
    res[i] = *--s *w;           /* compute average/expected values */
  return res;                   /* and return the result vector */
}  /* mat_mean() */

/*--------------------------------------------------------------------*/

double mat_isovar (const MATRIX *mat, int mle)
{                               /* --- isotropic variance computation */
  DIMID  i;                     /* loop variable */
  double sum, t;                /* sum of squared distances, buffer */

  assert(mat                    /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  for (sum = mat->els[0][0], i = mat->rowcnt; --i > 0; )
    sum += mat->els[i][i];      /* compute sum of squared distances */
  if (!(sum > 0)) return 0;     /* and check it */
  t = (mle) ? mat->weight : mat->weight-1;
  return (t > 0) ? sum /(t *(double)mat->rowcnt) : 0;
}  /* mat_isovar() */          /* return the isotropic variance */

/*--------------------------------------------------------------------*/

double mat_isovarm (double *res, const MATRIX *mat, int mle)
{                               /* --- isotropic variance computation */
  DIMID  i;                     /* loop variable */
  double sum, t, x;             /* sum of squared distances, buffers */
  const double *s;              /* to traverse the source vector */

  assert(mat && res             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt));
  t = (mat->weight > 0) ? 1/mat->weight : 0;
  for (sum = 0, s = mat->vec +(i = mat->rowcnt); --i >= 0; ) {
    x = *--s; res[i] = x *t;    /* compute average/expected values */
    sum += mat->els[i][i] -x *res[i];
  }                             /* compute sum of squared distances */
  if (!(sum > 0)) return 0;     /* and check it */
  t = (mle) ? mat->weight : mat->weight-1;
  return (t > 0) ? sum /(t *(double)mat->rowcnt) : 0;
}  /* mat_isovarm() */           /* return the isotropic variance */

/*--------------------------------------------------------------------*/

MATRIX* mat_isovarx (MATRIX *res, const MATRIX *mat, int mle)
{ res->els[0][0] = mat_isovarm(res->vec, mat, mle); return res; }

/*--------------------------------------------------------------------*/

MATRIX* mat_var (MATRIX *res, const MATRIX *mat, int mle)
{                               /* --- variance computation */
  DIMID  row;                   /* loop variable */
  double t, v;                  /* temporary buffers */

  assert(mat && res             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == res->colcnt)
  &&    (mat->rowcnt == res->colcnt));
  t = (mle) ? mat->weight : mat->weight-1;
  t = (t > 0) ? 1/t : 0;        /* get factor for variance estimation */
  for (row = res->rowcnt; --row >= 0; ) {
    v = mat->els[row][row] *t;  /* compute the variance */
    res->els[row][row] = (v > 0) ? v : 0;
  }                             /* store the variances and */
  return res;                   /* return the result matrix */
}  /* mat_var() */

/*--------------------------------------------------------------------*/

MATRIX* mat_varx (MATRIX *res, const MATRIX *mat, int mle)
{                               /* --- variance computation */
  DIMID  row;                   /* loop variable */
  double *b;                    /* to traverse the buffer */
  const double *s;              /* dito, for the source matrix */
  double w, t, v;               /* temporary buffers */

  assert(mat && res             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == res->colcnt)
  &&    (mat->rowcnt == res->colcnt));
  w = mat->weight;              /* get the weight sum */
  t = (mle)   ? w : w-1;        /* compute the factors for */
  t = (t > 0) ? 1/t : 0;        /* the variance estimation */
  w = (w > 0) ? 1/w : 0;        /* and the mean estimation */
  b = res->vec +res->rowcnt;    /* traverse the sum of the weighted */
  s = mat->vec +res->rowcnt;    /* vectors (stored in the buffer) */
  for (row = res->rowcnt; --row >= 0; ) {
    v = *--s; *--b = v *w;      /* compute average/expected values */
    v = t *(mat->els[row][row] -*b *v);
    res->els[row][row] = (v > 0) ? v : 0;
  }                             /* compute the variances and */
  return res;                   /* return the result matrix */
}  /* mat_varx() */

/*--------------------------------------------------------------------*/

MATRIX* mat_covar (MATRIX *res, const MATRIX *mat, int mle)
{                               /* --- covariance matrix computation */
  DIMID  row, col;              /* loop variables */
  double *d;                    /* to traverse the matrix rows */
  const double *s;              /* dito, for the source matrix */
  double t;                     /* temporary buffer */

  assert(mat && res             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == res->colcnt)
  &&    (mat->rowcnt == res->colcnt));
  t = (mle) ? mat->weight : mat->weight-1;
  t = (t > 0) ? 1/t : 0;        /* get factor for (co)variance estim. */
  for (row = res->rowcnt; --row >= 0; ) {
    s = mat->els[row];          /* traverse the lower matrix triangle */
    for (d = res->els[row] +(col = mat->colcnt); --col >= row; )
      *--d = s[col] *t;         /* compute the (co)variances and */
    if (!(*d > 0)) *d = 0;      /* ensure a non-negative diagonal */
  }
  return res;                   /* return the (covariance) matrix */
}  /* mat_covar() */

/*--------------------------------------------------------------------*/

MATRIX* mat_covarx (MATRIX *res, const MATRIX *mat, int mle)
{                               /* --- covariance matrix computation */
  DIMID  row, col;              /* loop variables */
  double *b, *d;                /* to traverse the buffer/matrix rows */
  const double *s;              /* dito, for the source matrix */
  double w, t, v;               /* temporary buffers */

  assert(mat && res             /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == res->colcnt)
  &&    (mat->rowcnt == res->colcnt));
  w = mat->weight;              /* get the weight sum */
  t = (w > 0) ? 1/w : 0;        /* and its reciprocal */
  b = res->vec +res->rowcnt;    /* traverse the sum of the weighted */
  s = mat->vec +res->rowcnt;    /* vectors (stored in the buffer) */
  for (row = res->rowcnt; --row >= 0; )
    *--b = *--s *t;             /* compute average/expected values */
  t = (mle)   ? w : w-1;        /* compute the factor for */
  t = (t > 0) ? 1/t : 0;        /* the (co)variance estimation */
  for (row = res->rowcnt; --row >= 0; ) {
    v = w *b[row];              /* traverse the element product sums */
    s = mat->els[row];          /* in the lower matrix triangle */
    for (d = res->els[row] +(col = mat->colcnt); --col >= row; )
      *--d = t *(s[col] -b[col] *v);
    if (!(*d > 0)) *d = 0;      /* compute the (co)variances and */
  }                             /* ensure a non-negative diagonal */
  return res;                   /* return the (covariance) matrix */
}  /* mat_covarx() */

/*--------------------------------------------------------------------*/

MATRIX* mat_correl (MATRIX *res, const MATRIX *mat)
{                               /* --- correlation from covar. matrix */
  DIMID  row, col;              /* loop variables */
  double *d; const double *s;   /* to traverse the matrix rows */
  double t;                     /* temporary buffer */

  assert(res && mat             /* check the function arguments */
  &&    (res->rowcnt == res->colcnt)
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt));
  for (row = res->rowcnt; --row >= 0; ) {
    t = mat->els[row][row];     /* traverse the variances (diagonal) */
    t = (t > 0) ? sqrt(t) : 0;  /* and compute their square roots */
    res->buf[row] = (t > 0) ? 1/t : 0;
  }                             /* note the standard deviations */
  for (row = res->rowcnt; --row >= 0; ) {
    t = res->buf[row];          /* get next needed std. deviation */
    d = res->els[row];          /* traverse the matrix rows */
    s = mat->els[row];          /* of source and destination */
    for (col = res->colcnt; --col > row; )
      d[col] = s[col] *res->buf[col] *t;
  }                             /* compute the correlation coeff. */
  for (row = res->rowcnt; --row >= 0; )
    res->els[row][row] = 1;     /* set the diagonal to 1 */
  return res;                   /* return the correlation matrix */
}  /* mat_correl() */           /* (only upper triangle is computed) */

/*--------------------------------------------------------------------*/

double mat_regress (MATRIX *res, double *rhs, const MATRIX *mat)
{                               /* --- set up regression eq. system */
  DIMID  row, col, n;           /* loop variables, number of columns */
  double *d; const double *s;   /* to traverse the matrix rows */
  double *b;                    /* to traverse the buffer */
  double so2;                   /* sum of squared outputs */

  assert(rhs && res && mat      /* check the function arguments */
  &&    (res->rowcnt == res->colcnt)
  &&    (mat->rowcnt == mat->colcnt)
  &&    (res->rowcnt == mat->rowcnt));
  rhs += n = mat->colcnt;       /* copy the sum of output values */
  *--rhs = mat->vec[--n];       /* to the right hand side vector */
  for (b = res->vec +(row = n); --row >= 0; ) {
    *--rhs = mat->els[row][n];  /* move the last column to the rhs */
    res->els[row][n] = *--b;    /* and the buffer to the last column */
  }                             /* (build system of normal equations) */
  so2 = mat->els[n][n];         /* note the sum of squared outputs */
  res->els[n][n] = mat->weight; /* store the sum of vector weights */
  if (res != mat) {             /* if not to work in place, */
    res->weight = mat->weight;  /* copy the sum of vector weights */
    for (row = n; --row >= 0; ) {
      s = mat->els[row] +n;     /* copy the upper triangle */
      d = res->els[row] +n;     /* of the aggregation matrix */
      for (col = n; --col >= row; ) *--d = *--s;
    }                           /* (only the upper triangle is needed */
  }                             /*  for the Cholesky decomposition) */
  return so2;                   /* return the sum of squared outputs */
}  /* mat_regress() */
